@@ -1,7 +1,6 @@
 import os
 import sys
 import json
-import time
 import queue
 import logging
 import readline
@@ -11,6 +10,8 @@ import threading
 from collections import OrderedDict
 
 from flask import Flask, jsonify, request, render_template
+from flask.helpers import make_response
+from flask.json import JSONEncoder
 from apscheduler.schedulers.blocking import BlockingScheduler
 
 from watchmen.listener import (
@@ -44,6 +45,22 @@ gpu_queue.put(GPUInfo())
 client_queue = queue.Queue()
 client_queue.put(ClientCollection())
 
+APP_PORT = None
+
+
+class CustomJSONEncoder(JSONEncoder):
+    def default(self, obj):
+        try:
+            if isinstance(obj, datetime.datetime):
+                return obj.strftime("%Y-%m-%d %H:%M:%S")
+            iterable = iter(obj)
+        except TypeError:
+            pass
+        else:
+            return list(iterable)
+        return JSONEncoder.default(self, obj)
+app.json_encoder = CustomJSONEncoder
+
 
 @app.route("/gpu/<int:gpu_id>")
 def get_single_gpu_status(gpu_id: int):
@@ -55,7 +72,7 @@ def get_single_gpu_status(gpu_id: int):
     except ValueError as err:
         msg = str(err)
         status = "err"
-    return jsonify(**{"status": status, "msg": msg})
+    return jsonify({"status": status, "msg": msg})
 
 
 @app.route("/gpus/<gpu_ids>")
@@ -78,7 +95,7 @@ def get_gpus_status(gpu_ids: str):
     except ValueError as err:
         msg = str(err)
         status = "err"
-    return jsonify(**{"status": status, "msg": msg, "detail": detail})
+    return jsonify({"status": status, "msg": msg, "detail": detail})
 
 
 @app.route("/client/ping", methods=["POST"])
@@ -98,9 +115,9 @@ def client_ping():
         status = "err"
         msg = "cannot ping before register"
     client_queue.put(cc)
-    return jsonify(**{"status": status, 
-                      "available_gpus": available_gpus,
-                      "msg": msg})
+    return jsonify({"status": status,
+                    "available_gpus": available_gpus,
+                    "msg": msg})
 
 
 @app.route("/client/register", methods=["POST"])
@@ -116,7 +133,7 @@ def client_register():
         msg = f"mode {client_info.mode} is not supported"
     elif not check_gpus_existence(client_info.gpus):
         status = "err"
-        msg = f"check the gpus existence"
+        msg = "check the gpus existence"
     elif client_info.mode == ClientMode.SCHEDULE and \
             not check_req_gpu_num(client_info.req_gpu_num):
         status = "err"
@@ -128,6 +145,7 @@ def client_register():
                 id=client_info.id,
                 mode=client_info.mode,
                 status=ClientStatus.WAITING,
+                register_time=datetime.datetime.now(),
                 last_request_time=datetime.datetime.now(),
                 queue_num=len(cc.work_queue),
                 gpus=client_info.gpus,
@@ -139,7 +157,7 @@ def client_register():
             status = "err"
             msg = f"client_id: {client_info.id} has been registered!"
         client_queue.put(cc)
-    return jsonify(**{"status": status, "msg": msg})
+    return jsonify({"status": status, "msg": msg})
 
 
 @app.route("/show/work", methods=["GET"])
@@ -155,7 +173,7 @@ def show_work():
         msg = str(err)
     finally:
         client_queue.put(cc)
-    return jsonify(**{"status": status, "msg": msg})
+    return jsonify({"status": status, "msg": msg})
 
 
 @app.route("/show/finished", methods=["GET"])
@@ -171,7 +189,7 @@ def show_finished():
         msg = str(err)
     finally:
         client_queue.put(cc)
-    return jsonify(**{"status": status, "msg": msg})
+    return jsonify({"status": status, "msg": msg})
 
 
 @app.route("/show/gpus", methods=["GET"])
@@ -188,18 +206,44 @@ def show_gpus():
         msg = str(err)
     finally:
         gpu_queue.put(gpu_info)
-    return jsonify(**{"status": status, "msg": msg})
+    return jsonify({"status": status, "msg": msg})
+
+
+@app.route("/api", methods=["GET", "OPTIONS"])
+def api():
+    if request.method == "OPTIONS":
+        response = make_response()
+    else:
+        gpu_info = show_gpus()
+        gpu_msg = gpu_info.json["msg"]
+        work_info = show_work()
+        work_msg = work_info.json["msg"]
+        finished_info = show_finished()
+        finished_msg = finished_info.json["msg"]
+        response = jsonify({
+            "gpu": gpu_msg,
+            "work_queue": work_msg,
+            "finished_queue": finished_msg
+        })
+    response.headers["Access-Control-Allow-Origin"] = "*"
+    return response
 
 
 @app.route("/", methods=["GET"])
 def index():
+    global APP_PORT
+    return render_template("index.html", port=APP_PORT)
+
+
+@app.route("/old", methods=["GET"])
+def old_index():
     gpu_info = show_gpus()
     gpu_msg = gpu_info.json["msg"]
     work_info = show_work()
     work_msg = work_info.json
     finished_info = show_finished()
     finished_msg = finished_info.json
-    return render_template("index.html",
+    return render_template("old_index.html",
                            gpu_msg=gpu_msg,
                            work_msg=work_msg,
                            finished_msg=finished_msg)
@@ -214,18 +258,19 @@ def check_gpu_info():
 def check_work(queue_timeout):
     cc = client_queue.get()
     marked_finished = []
-    reserved_gpus = set() # whether there can be multiple `ok` in one scan
+    reserved_gpus = set()  # whether there can be multiple `ok` in one scan
     queue_num = 0
     for client_id, client in cc.work_queue.items():
         time_delta = datetime.datetime.now() - client.last_request_time
         if time_delta.seconds > queue_timeout:
             if client.status != ClientStatus.OK:
                 client.status = ClientStatus.TIMEOUT
-            client.queue_num = -1 # invalid client
+            client.queue_num = -1  # invalid client
             marked_finished.append(client_id)
             continue
         client.queue_num = queue_num
         ok = False
+        available_gpus = []
         if client.status == ClientStatus.OK:
             reserved_gpus |= set(client.gpus)
         else:
@@ -236,7 +281,7 @@ def check_work(queue_timeout):
                     available_gpus = client.gpus
                 elif client.mode == 'schedule':
                     ok, available_gpus = gpu_info.is_req_gpu_num_satisfied(
-                            client.gpus, client.req_gpu_num)
+                        client.gpus, client.req_gpu_num)
                 else:
                     raise RuntimeError(f"Not supported mode: {client.mode}")
             except IndexError as err:
@@ -262,9 +307,8 @@ def check_finished(status_queue_keep_time):
     cc = client_queue.get()
     marked_delete_ids = []
     for client_id, client in cc.finished_queue.items():
-        time_delta = datetime.datetime.now() - client.last_request_time
-        # there is no timedelta.hours attribute and 1 hour = 3600 seconds
-        if time_delta.seconds/3600 >= status_queue_keep_time:
+        delta = datetime.datetime.now() - client.last_request_time
+        if (delta.days * 24 + delta.seconds / 3600) >= status_queue_keep_time:
             marked_delete_ids.append(client_id)
     for client_id in marked_delete_ids:
         cc.finished_queue.pop(client_id)
@@ -279,18 +323,21 @@ def regular_check(request_interval, queue_timeout, status_queue_keep_time):
                       next_run_time=datetime.datetime.now())
     scheduler.add_job(check_work,
                       trigger='interval',
-                      seconds=request_interval*5,
+                      seconds=request_interval * 5,
                       args=(queue_timeout,),
                       next_run_time=datetime.datetime.now())
-    scheduler.add_job(check_finished,
-                      trigger='interval',
-                      hours=status_queue_keep_time,
-                      args=(status_queue_keep_time,),
-                      next_run_time=datetime.datetime.now())
+    if status_queue_keep_time != -1:
+        scheduler.add_job(check_finished,
+                        trigger='interval',
+                        hours=status_queue_keep_time,
+                        args=(status_queue_keep_time,),
+                        next_run_time=datetime.datetime.now())
     scheduler.start()
 
 
 def api_server(host, port):
+    global APP_PORT
+    APP_PORT = port
     app.run(host=host, port=port)
 
 
@@ -305,9 +352,10 @@ if __name__ == "__main__":
     parser.add_argument("--request_interval", type=int, default=1,
                         help="interval for gpu status requesting (seconds)")
     parser.add_argument("--status_queue_keep_time", type=int, default=48,
-                        help="hours for keeping the client status")
+                        help=("hours for keeping the client status. "
+                              "set `-1` to keep all clients' status"))
     args = parser.parse_args()
-    
+
     logger.info(f"Running at: {args.host}:{args.port}")
     logger.info(f"Current pid: {os.getpid()} > watchmen_server.pid")
     with open("watchmen_server.pid", "wt", encoding="utf-8") as fout:
@@ -341,7 +389,7 @@ if __name__ == "__main__":
             if not api_server_worker.is_alive():
                 logger.error("api server worker is not alive, server quit")
                 raise RuntimeError("api server worker is not alive, server quit")
-        except RuntimeError as err:
+        except RuntimeError:
             logger.error("runtime error, kill the server")
             break
         except KeyboardInterrupt:
